import pytest
import struct

from ofrak_type.endianness import Endianness


@pytest.mark.parametrize(
    "endianness,expected_struct_flag",
    [
        (Endianness.BIG_ENDIAN, ">"),
        (Endianness.LITTLE_ENDIAN, "<"),
    ],
)
def test_endianness_get_struct_flag(endianness: Endianness, expected_struct_flag: str):
    assert endianness.get_struct_flag() == expected_struct_flag

    # Test the struct flag actually works with struct module
    test_value = 0x1234
    packed = struct.pack(f"{expected_struct_flag}H", test_value)
    unpacked = struct.unpack(f"{expected_struct_flag}H", packed)[0]
    assert unpacked == test_value

    # Verify endianness behavior
    if endianness == Endianness.BIG_ENDIAN:
        assert packed == b"\x12\x34"
    elif endianness == Endianness.LITTLE_ENDIAN:
        assert packed == b"\x34\x12"



from enum import Enum


class MemoryPermissions(Enum):
    """
    Representation of memory access permissions - all combinations of Read, Write, and eXecute
    are represented.
    """

    NONE = 0
    X = 1
    W = 2
    R = 4
    RW = R + W
    RX = R + X
    WX = W + X
    RWX = R + W + X

    def as_str(self) -> str:
        if self.value == 0:
            return "none"

        string = ""
        if self.value & MemoryPermissions.R.value:
            string += "r"
        if self.value & MemoryPermissions.W.value:
            string += "w"
        if self.value & MemoryPermissions.X.value:
            string += "x"
        return string

    def __add__(self, other: "MemoryPermissions") -> "MemoryPermissions":
        if not isinstance(other, MemoryPermissions):
            raise TypeError(f"Operation between MemoryPermissions and {type(other)} not supported")
        elif other.value & self.value != 0:
            raise ValueError(f"Cannot add {self} and {other} because they overlap!")
        else:
            return MemoryPermissions(self.value + other.value)

    def __and__(self, other: "MemoryPermissions") -> "MemoryPermissions":
        if not isinstance(other, MemoryPermissions):
            raise TypeError(f"Operation between MemoryPermissions and {type(other)} not supported")
        else:
            return MemoryPermissions(self.value & other.value)

    def __sub__(self, other: "MemoryPermissions") -> "MemoryPermissions":
        if not isinstance(other, MemoryPermissions):
            raise TypeError(f"Operation between MemoryPermissions and {type(other)} not supported")
        elif other.value == 0:
            return self
        elif other.value & self.value == 0:
            raise ValueError(f"Cannot subtract {other} from {self} because they have no overlap!")
        else:
            return MemoryPermissions(self.value & (~other.value & 0x7))



from enum import Enum


class BitWidth(Enum):
    """
    The number of bits which can be used to represent a number.
    """

    BIT_8 = 8
    BIT_16 = 16
    BIT_32 = 32
    BIT_64 = 64

    def get_word_size(self):  # type: () -> int
        return int(self.value / 8)



from enum import Enum
from typing import Optional

from dataclasses import dataclass

from ofrak_type.bit_width import BitWidth
from ofrak_type.endianness import Endianness


class InstructionSet(Enum):
    """
    Enumeration of the possible supported instruction sets.

    :ivar ARM: ARM
    :ivar AARCH64: ARM 64-bit
    :ivar X86: Intel x86
    :ivar MIPS: MIPS
    :ivar MSP430: MSP430
    :ivar PPS: PowerPC
    :ivar M68K: Motorola 68K
    :ivar AVR: Atmel AVR
    :ivar SPARC: SPARC
    """

    ARM = "arm"
    AARCH64 = "aarch64"
    X86 = "x86"
    MIPS = "mips"
    MSP430 = "msp430"
    PPC = "ppc"
    M68K = "68000"
    AVR = "avr"
    SPARC = "sparc"


class SubInstructionSet(Enum):
    """
    ARM sub instruction set. Different sub ISAs correspond to different
    [ARM versions](https://en.wikipedia.org/wiki/AArch64).
    """

    ARMv4T = "ARMV4T"
    ARMv5T = "ARMV5T"
    ARMv5TE = "ARMV5TE"
    ARMv6 = "ARMV6"
    ARMv6J = "ARMV6J"
    ARMv6K = "ARMV6K"
    ARMv6KZ = "ARMV6KZ"
    ARMv6T2 = "ARMV6T2"
    ARMv6Z = "ARMV6Z"
    ARMv6ZK = "ARMV6ZK"
    ARMv7 = "ARMV7"
    ARMv7A = "ARMV7-A"
    ARMv7VE = "ARMV7VE"
    ARMv8A = "ARMV8-A"
    ARMv81A = "ARMV8.1-A"
    ARMv82A = "ARMV8.2-A"
    ARMv83A = "ARMV8.3-A"
    ARMv84A = "ARMV8.4-A"
    ARMv85A = "ARMV8.5-A"
    ARMv86A = "ARMV8.6-A"
    ARMv7R = "ARMV7-R"
    ARMv8R = "ARMV8-R"
    ARMv6M = "ARMV6-M"
    ARMv6SM = "ARMV6S-M"
    ARMv7M = "ARMV7-M"
    ARMv7EM = "ARMV7E-M"
    ARMv8MBASE = "ARMV8-M.BASE"
    ARMv8MMAIN = "ARMV8-M.MAIN"
    ARMv81MMAIN = "ARMV8.1-M.MAIN"
    ARMv9A = "ARMV9-A"
    IWMMXT = "IWMMXT"
    IWMMXT2 = "IWMMXT2"
    # AVR
    AVR2 = "avr2"
    AVR25 = "avr25"
    AVR3 = "avr3"
    AVR31 = "avr31"
    AVR35 = "avr35"
    AVR4 = "avr4"
    AVR5 = "avr5"
    AVR51 = "avr51"
    AVR6 = "avr6"
    AVRXMEGA2 = "avrxmega2"
    # AVRXMEGA3 = "avrxmega3"  # Not supported in 5.4.0
    AVRXMEGA4 = "avrxmega4"
    AVRXMEGA5 = "avrxmega5"
    AVRXMEGA6 = "avrxmega6"
    AVRXMEGA7 = "avrxmega7"
    AVRTINY = "avrtiny"
    AVR1 = "avr1"  # assembler only


class InstructionSetMode(Enum):
    """
    Instruction set mode. Useful for architectures which have multiple encodings it can switch
    between on the fly. In particular the Thumb mode for ARM and VLE more for PPC are represented.

    :ivar NONE: None
    :ivar THUMB: Thumb (ARM)
    :ivar VLE: VLE (PPC)
    """

    NONE = 0
    THUMB = 1
    VLE = 2


class ProcessorType(Enum):
    """
    Enumeration of specific processor types.

    :ivar ARM926EJ_S:
    :ivar GENERIC_A53_V8:
    :ivar GENERIC_A9_V6:
    :ivar GENERIC_A9_V7:
    :ivar GENERIC_A9_V7_THUMB:
    :ivar MSP430:
    :ivar MIPS_LITTLE:
    :ivar MIPS_RM5721_BIG:
    :ivar PPC_405:
    :ivar PPC_MPC855T:
    :ivar PPC_VLE:
    :ivar PI_ARM6:
    :ivar PIXHAWK:
    :ivar SPARTAN:
    :ivar X186_16:
    :ivar X186_32:
    :ivar X64:
    :ivar XSCALE:
    :ivar COLDFIRE4E:
    :ivar CORTEX_A53:
    :ivar AVR:
    :ivar SPARC:
    """

    ARM926EJ_S = "arm926ej-s"
    GENERIC_A53_V8 = "generic_a53_v8"
    GENERIC_A9_V6 = "generic_a9_v6"
    GENERIC_A9_V7 = "generic_a9_v7"
    GENERIC_A9_V7_THUMB = "generic_a9_v7_thumb"
    GENERIC_ARM_BE8 = "generic_be8"  # ARM with little endian code but big endian data
    MSP430 = "msp340"
    MIPS_LITTLE = "mips"
    MIPS_RM5721_BIG = "mips_rm5721_big"
    PPC_405 = "ppc_405"
    PPC_MPC855T = "ppc_mpc855t"
    PPC_VLE = "ppc_vle"
    PI_ARM6 = "pi_arm6"
    PIXHAWK = "pixhawk"
    SPARTAN = "spartan"
    X186_16 = "x186_16"
    X186_32 = "x186_32"
    I386 = "i386"
    X64 = "x86_64"
    XSCALE = "xscale"
    COLDFIRE4E = "cfv4e"
    CORTEX_A53 = "cortex-a53"
    CORTEX_A55 = "cortex-a55"
    CORTEX_A72 = "cortex-a72"
    AVR = "avr"
    SPARC = "leon3"


@dataclass(frozen=True, eq=True)
class ArchInfo:
    """
    Collection of fields used to describe an architecture

    :ivar isa: Instruction set architecture
    :ivar sub_isa: Sub instruction set
    :ivar bit_width: Bits per word
    :ivar endianness: Endianness as `Endianness.BIG_ENDIAN` or `Endianness.LITTLE_ENDIAN`
    :ivar processor: Processor type
    """

    isa: InstructionSet
    sub_isa: Optional[SubInstructionSet]
    bit_width: BitWidth
    endianness: Endianness
    processor: Optional[ProcessorType]



import pytest

from ofrak_type.bit_width import BitWidth


@pytest.mark.parametrize(
    "bitwidth,expected_word_size",
    [(BitWidth.BIT_8, 1), (BitWidth.BIT_16, 2), (BitWidth.BIT_32, 4), (BitWidth.BIT_64, 8)],
)
def test_bitwidth_get_word_size(bitwidth: BitWidth, expected_word_size: int):
    assert bitwidth.get_word_size() == expected_word_size



from ofrak_type.architecture import *
from ofrak_type.bit_width import *
from ofrak_type.endianness import *
from ofrak_type.error import *
from ofrak_type.memory_permissions import *
from ofrak_type.range import *
from ofrak_type.symbol_type import *



from dataclasses import dataclass

import sys
from typing import Iterable, List


@dataclass
class Range:
    """
    Range of addresses.

    :ivar start: The start address of the Range
    :ivar end: The end address of the Range
    """

    start: int
    end: int

    MAX = sys.maxsize

    def __post_init__(self):
        if self.start > self.end:
            raise ValueError("The start value must be less than or equal to the end value")

    def __iter__(self):
        """
        Make this class iterable, allowing to write `for i in Range(...)` and `if i in Range(...)`.
        """
        yield from range(self.start, self.end)

    def __contains__(self, value):
        """
        Make the check `value in Range(...)` constant time.
        """
        return self.contains_value(value)

    def length(self):
        """
        Compute the length of this range
        """
        return self.end - self.start

    def contains_value(self, value: int) -> bool:
        """
        Determine if the provided value is within the range (inclusive of the start, exclusive of
        the end).

        Note: you can use `value in r` alternatively to `r.contains_value(value)`.
        """
        return self.start <= value < self.end

    def within(self, range: "Range") -> bool:
        """
        Determine if this range is within the provided range
        """
        return self.start >= range.start and self.end <= range.end

    def overlaps(self, range: "Range") -> bool:
        """
        Determine if this range overlaps the provided range.
        """
        return range.start < self.end and range.end > self.start

    def intersect(self, range: "Range") -> "Range":
        """
        Compute the largest possible range that is within both this range and the provided range.
        It raises a ValueError if no such range exists.
        """
        start = max(range.start, self.start)
        end = min(range.end, self.end)
        if start > end:
            raise ValueError("There is no overlap between this range and the provided range ")
        return Range(start, end)

    def split(self, range: "Range") -> Iterable["Range"]:
        """
        Split the range into one or more ranges that do not overlap the provided range
        :param range:
        :return:
        """
        if range.start >= self.end or range.end <= self.start:
            # No overlap
            return (self,)
        elif self.within(range):
            # This range is covered by the provided range
            return tuple()
        elif range.start <= self.start and range.end < self.end:
            # Overlap on the right
            return (Range(range.end, self.end),)
        elif range.start > self.start and range.end >= self.end:
            # Overlap on the left
            return (Range(self.start, range.start),)
        elif range.within(self):
            # The provided range sits in the middle of this range
            return Range(self.start, range.start), Range(range.end, self.end)
        else:
            raise ValueError("Unreachable")

    def translate(self, offset: int) -> "Range":
        """
        Generate a new range based on this range translated by the provided offset
        """
        if offset == 0:
            return self
        if offset + self.start < 0:
            raise ValueError("The start of the translated range cannot be negative")

        if self.end + offset > Range.MAX:
            raise OverflowError("Translated range exceeds maximum allowed value")

        return Range(self.start + offset, self.end + offset)

    def __repr__(self) -> str:
        return f"Range({hex(self.start)}, {hex(self.end)})"

    def __hash__(self):
        return hash((self.start, self.end))

    @staticmethod
    def from_size(start: int, size: int) -> "Range":
        return Range(start, start + size)

    @staticmethod
    def merge_ranges(ranges: Iterable["Range"]) -> List["Range"]:
        """
        Merge multiple Ranges into a minimal set of Ranges. The algorithm here is basically
        finding where the input Ranges do NOT overlap, and inverting that.

        In more detail the algorithm is essentially:
        1. Iterates over all values from the minimum start to the maximum end while tracking a
        counter
        2. Increment the counter whenever the start of a range is reached, decrement the counter
        when the end of a range is reached
        3. Save a merged range when the counter is 0.

        It's more efficient than that since only range starts/ends are iterated over, but that's
        the idea.

        :param ranges: unordered iterable of Range objects to merge
        """
        range_bounds_markers = []
        for r in ranges:
            range_bounds_markers.append((r.start, 1))
            range_bounds_markers.append((r.end, -1))

        # Sort markers by index, and use inverted second item in tuple (incr) as tiebreaker
        # If start and end markers have the same index, the start marker(s) should be counted first
        range_bounds_markers.sort(key=lambda idx_incr: (idx_incr[0], -idx_incr[1]))

        merged_ranges = []
        current_overlapping_ranges = 0
        last_range_start = 0
        for idx, incr in range_bounds_markers:
            if current_overlapping_ranges == 0:
                # Must be at the start of a range
                last_range_start = idx
            current_overlapping_ranges += incr
            # The counter should never drop below zero
            assert current_overlapping_ranges >= 0, "Range bounds markers not ordered correctly"
            if current_overlapping_ranges == 0:
                # Must be at the end of a range
                merged_ranges.append(Range(last_range_start, idx))

        return merged_ranges


def chunk_ranges(ranges: List[Range], chunk_size: int) -> List[Range]:
    """
    Break a list of Ranges into equal sized regions of Ranges, assuming each range is evenly
    divisible by chunk_size.

    :param ranges:
    :param chunk_size:
    :return: equal sized regions of Ranges
    """
    if chunk_size <= 0:
        raise ValueError("chunk_size must be positive")

    regions = Range.merge_ranges(ranges)
    chunked = []
    for region in regions:
        for i in range(region.start, region.end, chunk_size):
            chunked.append(Range(i, min(i + chunk_size, region.end)))
    return chunked


def remove_subranges(ranges: List[Range], to_remove: List[Range]) -> List[Range]:
    """
    Subtract one set of addresses from another, both expressed as a list of non-overlapping ranges.

    :param ranges: A list of non-overlapping ranges.
    :param to_remove: A list of non-overlapping ranges to be removed from the first argument.
    :return: A list of ranges covering the input ranges with the subranges removed.
    """
    if not ranges or not to_remove:
        return ranges

    ranges.sort(key=lambda range_: range_.start)
    to_remove.sort(key=lambda range_: range_.start)

    i = 0
    j = 0
    ret = []
    current_range = ranges[i]
    current_to_remove = to_remove[j]
    while True:
        if current_range.start >= current_to_remove.end:
            j += 1
            if j < len(to_remove):
                current_to_remove = to_remove[j]
            else:
                ret.append(current_range)
                ret.extend(ranges[i + 1 :])
                break
        elif current_range.end <= current_to_remove.start:
            ret.append(current_range)
            i += 1
            if i < len(ranges):
                current_range = ranges[i]
            else:
                break

        elif current_range.start < current_to_remove.start:
            ret.append(Range(current_range.start, current_to_remove.start))
            current_range = Range(current_to_remove.start, current_range.end)

        elif current_range.end > current_to_remove.end:
            current_range = Range(current_to_remove.end, current_range.end)

        else:
            # current_to_remove contains current_range
            i += 1
            if i >= len(ranges):
                break
            current_range = ranges[i]

    return ret



from enum import Enum


class Endianness(Enum):
    """
    The order in which bytes are stored.
    """

    BIG_ENDIAN = "big"
    LITTLE_ENDIAN = "little"

    def get_struct_flag(self) -> str:
        if self is Endianness.BIG_ENDIAN:
            return ">"
        else:
            return "<"



import pytest

from ofrak_type.memory_permissions import MemoryPermissions


@pytest.mark.parametrize(
    "memory_permissions, expected_str",
    [
        (MemoryPermissions.R, "r"),
        (MemoryPermissions.W, "w"),
        (MemoryPermissions.X, "x"),
        (MemoryPermissions.RW, "rw"),
        (MemoryPermissions.RX, "rx"),
        (MemoryPermissions.WX, "wx"),
        (MemoryPermissions.RWX, "rwx"),
        (MemoryPermissions.NONE, "none"),
    ],
)
def test_memory_permissions_as_str(memory_permissions: MemoryPermissions, expected_str: str):
    assert memory_permissions.as_str() == expected_str


class TestMemoryPermissionsAdd:
    def test_add(self):
        # Basic additions
        assert MemoryPermissions.R + MemoryPermissions.X == MemoryPermissions.RX
        assert MemoryPermissions.R + MemoryPermissions.W == MemoryPermissions.RW
        assert MemoryPermissions.W + MemoryPermissions.X == MemoryPermissions.WX

        # Adding NONE should return the original
        assert MemoryPermissions.R + MemoryPermissions.NONE == MemoryPermissions.R
        assert MemoryPermissions.W + MemoryPermissions.NONE == MemoryPermissions.W
        assert MemoryPermissions.X + MemoryPermissions.NONE == MemoryPermissions.X
        assert MemoryPermissions.NONE + MemoryPermissions.NONE == MemoryPermissions.NONE

        # Building up to RWX
        assert MemoryPermissions.R + MemoryPermissions.WX == MemoryPermissions.RWX
        assert MemoryPermissions.W + MemoryPermissions.RX == MemoryPermissions.RWX
        assert MemoryPermissions.X + MemoryPermissions.RW == MemoryPermissions.RWX

    def test_add_type_error(self):
        with pytest.raises(TypeError):
            MemoryPermissions.R + 2  # type: ignore

    def test_add_value_error(self):
        with pytest.raises(ValueError):
            MemoryPermissions.R + MemoryPermissions.RWX  # type: ignore


class TestMemoryPermissionsAnd:
    def test_and(self):
        # Basic AND operations
        assert MemoryPermissions.RWX & MemoryPermissions.W == MemoryPermissions.W
        assert MemoryPermissions.RWX & MemoryPermissions.R == MemoryPermissions.R
        assert MemoryPermissions.RWX & MemoryPermissions.X == MemoryPermissions.X

        # AND with NONE always returns NONE
        assert MemoryPermissions.RWX & MemoryPermissions.NONE == MemoryPermissions.NONE
        assert MemoryPermissions.R & MemoryPermissions.NONE == MemoryPermissions.NONE
        assert MemoryPermissions.NONE & MemoryPermissions.NONE == MemoryPermissions.NONE

        # AND with itself returns itself
        assert MemoryPermissions.R & MemoryPermissions.R == MemoryPermissions.R
        assert MemoryPermissions.RW & MemoryPermissions.RW == MemoryPermissions.RW
        assert MemoryPermissions.RWX & MemoryPermissions.RWX == MemoryPermissions.RWX

        # Partial overlaps
        assert MemoryPermissions.RW & MemoryPermissions.RX == MemoryPermissions.R
        assert MemoryPermissions.RW & MemoryPermissions.WX == MemoryPermissions.W
        assert MemoryPermissions.RX & MemoryPermissions.WX == MemoryPermissions.X

        # No overlap returns NONE
        assert MemoryPermissions.R & MemoryPermissions.W == MemoryPermissions.NONE
        assert MemoryPermissions.R & MemoryPermissions.X == MemoryPermissions.NONE
        assert MemoryPermissions.W & MemoryPermissions.X == MemoryPermissions.NONE

    def test_and_type_error(self):
        with pytest.raises(TypeError):
            MemoryPermissions.R & 2  # type: ignore


class TestMemoryPermissionsSub:
    def test_sub(self):
        # Basic subtractions
        assert MemoryPermissions.RWX - MemoryPermissions.W == MemoryPermissions.RX
        assert MemoryPermissions.RWX - MemoryPermissions.R == MemoryPermissions.WX
        assert MemoryPermissions.RWX - MemoryPermissions.X == MemoryPermissions.RW

        # Subtracting NONE returns the original
        assert MemoryPermissions.RWX - MemoryPermissions.NONE == MemoryPermissions.RWX
        assert MemoryPermissions.R - MemoryPermissions.NONE == MemoryPermissions.R
        assert MemoryPermissions.NONE - MemoryPermissions.NONE == MemoryPermissions.NONE

        # Subtracting multiple permissions
        assert MemoryPermissions.RWX - MemoryPermissions.RW == MemoryPermissions.X
        assert MemoryPermissions.RWX - MemoryPermissions.RX == MemoryPermissions.W
        assert MemoryPermissions.RWX - MemoryPermissions.WX == MemoryPermissions.R

        # Subtracting all permissions
        assert MemoryPermissions.R - MemoryPermissions.R == MemoryPermissions.NONE
        assert MemoryPermissions.RW - MemoryPermissions.RW == MemoryPermissions.NONE
        assert MemoryPermissions.RWX - MemoryPermissions.RWX == MemoryPermissions.NONE

        # Partial subtraction
        assert MemoryPermissions.RW - MemoryPermissions.R == MemoryPermissions.W
        assert MemoryPermissions.RW - MemoryPermissions.W == MemoryPermissions.R

    def test_sub_type_error(self):
        with pytest.raises(TypeError):
            MemoryPermissions.R - 0  # type: ignore

    def test_sub_value_error(self):
        with pytest.raises(ValueError):
            MemoryPermissions.RW - MemoryPermissions.X  # type: ignore
            MemoryPermissions.NONE - MemoryPermissions.R  # type: ignore


class TestMemoryPermissionsEnumValues:
    """Test that the enum values are correct"""

    def test_enum_values(self):
        assert MemoryPermissions.NONE.value == 0
        assert MemoryPermissions.X.value == 1
        assert MemoryPermissions.W.value == 2
        assert MemoryPermissions.R.value == 4
        assert MemoryPermissions.RW.value == 6  # 4 + 2
        assert MemoryPermissions.RX.value == 5  # 4 + 1
        assert MemoryPermissions.WX.value == 3  # 2 + 1
        assert MemoryPermissions.RWX.value == 7  # 4 + 2 + 1


class TestMemoryPermissionsProperties:
    """Test mathematical properties of operations"""

    def test_inverse_operations(self):
        # Adding then subtracting should return original
        assert (
            MemoryPermissions.R + MemoryPermissions.W - MemoryPermissions.W == MemoryPermissions.R
        )
        assert (
            MemoryPermissions.W + MemoryPermissions.X - MemoryPermissions.X == MemoryPermissions.W
        )
        assert (
            MemoryPermissions.R + MemoryPermissions.X - MemoryPermissions.R == MemoryPermissions.X
        )

    def test_chained_operations(self):
        # Complex chained operations
        result = MemoryPermissions.NONE + MemoryPermissions.R
        result = result + MemoryPermissions.W
        result = result + MemoryPermissions.X
        assert result == MemoryPermissions.RWX

        # Subtract in reverse order
        result = result - MemoryPermissions.X
        result = result - MemoryPermissions.W
        result = result - MemoryPermissions.R
        assert result == MemoryPermissions.NONE



from dataclasses import dataclass
from typing import List, Iterable

import pytest
from hypothesis import strategies, given
from hypothesis.strategies import composite

from ofrak_type.range import Range, remove_subranges, chunk_ranges


@composite
def range_strategy(draw):
    start = draw(strategies.integers(min_value=0, max_value=2**32 - 1))
    end = draw(strategies.integers(min_value=start, max_value=2**32 - 1))
    return Range(start, end)


def test_invalid_range():
    with pytest.raises(ValueError):
        _ = Range(5, 0)


@given(r=range_strategy())
def test_length(r: Range):
    assert r.length() == r.end - r.start


@pytest.mark.parametrize(
    "r, value, expected_result",
    [
        (Range(0, 10), 0, True),
        (Range(0, 10), 5, True),
        (Range(0, 10), 10, False),
        (Range(0, 10), -1, False),
    ],
)
def test_contains(r: Range, value: int, expected_result: bool):
    assert r.contains_value(value) is expected_result
    # The `in` operator should also work:
    assert (value in r) is expected_result


@pytest.mark.parametrize(
    "range_1, range_2, expected_result",
    [
        (Range(2, 5), Range(0, 10), True),
        (Range(0, 10), Range(2, 5), False),
        (Range(5, 10), Range(0, 10), True),
        (Range(0, 2), Range(0, 10), True),
        (Range(0, 19), Range(5, 15), False),
    ],
)
def test_within(range_1: Range, range_2: Range, expected_result: bool):
    assert range_1.within(range_2) is expected_result


@pytest.mark.parametrize(
    "range_1, range_2, expected_result",
    [
        (Range(0, 10), Range(5, 15), True),
        (Range(0, 5), Range(5, 10), False),
        (Range(0, 10), Range(5, 7), True),
        (Range(5, 7), Range(0, 10), True),
    ],
)
def test_overlaps(range_1: Range, range_2: Range, expected_result: bool):
    assert range_1.overlaps(range_2) is expected_result


@pytest.mark.parametrize(
    "range_1, range_2, expected_range",
    [
        (Range(10, 20), Range(20, 30), Range(20, 20)),
        (Range(10, 20), Range(0, 30), Range(10, 20)),
        (Range(10, 20), Range(10, 30), Range(10, 20)),
        (Range(10, 20), Range(15, 30), Range(15, 20)),
        (Range(10, 20), Range(10, 20), Range(10, 20)),
    ],
)
def test_intersect(range_1: Range, range_2: Range, expected_range: Range):
    assert range_1.intersect(range_2) == expected_range


def test_intersect_value_error():
    range_1 = Range(10, 20)
    with pytest.raises(ValueError):
        range_1.intersect(Range(30, 40))


@pytest.mark.parametrize(
    "range_1, range_2, expected_result",
    [
        (Range(0, 10), Range(10, 20), (Range(0, 10),)),
        (Range(0, 1), Range(0, 1), ()),
        (Range(5, 10), Range(4, 9), (Range(9, 10),)),
        (Range(5, 10), Range(6, 11), (Range(5, 6),)),
        (
            Range(5, 10),
            Range(7, 8),
            (
                Range(5, 7),
                Range(8, 10),
            ),
        ),
        (
            Range(0, 10),
            Range(5, 5),
            (
                Range(0, 5),
                Range(5, 10),
            ),
        ),  # split with empty range
    ],
)
def test_split(range_1: Range, range_2: Range, expected_result: Iterable[Range]):
    assert range_1.split(range_2) == expected_result


def test_split_value_error():
    """
    Test that the unreachable else statement at the end of Range.split raises a ValueError.
    """

    class RangeNeverWithin(Range):
        def within(self, range: Range) -> bool:
            return False

    range_1 = Range(5, 10)
    range_2 = RangeNeverWithin(7, 8)
    with pytest.raises(ValueError):
        range_1.split(range_2)


REMOVE_SUBRANGES_TEST_CASES = [
    (
        [Range(0x20, 0x40), Range(0x60, 0x80)],
        [Range(0x30, 0x50)],
        [Range(0x20, 0x30), Range(0x60, 0x80)],
    ),
    (
        [Range(0x20, 0x40), Range(0x60, 0x80)],
        [Range(0x50, 0x70)],
        [Range(0x20, 0x40), Range(0x70, 0x80)],
    ),
    (
        [Range(0x20, 0x40), Range(0x60, 0x80)],
        [Range(0x30, 0x70)],
        [Range(0x20, 0x30), Range(0x70, 0x80)],
    ),
    (
        [Range(0x20, 0x40), Range(0x60, 0x80)],
        [Range(0x00, 0x20), Range(0x40, 0x60), Range(0x80, 0xA0)],
        [Range(0x20, 0x40), Range(0x60, 0x80)],
    ),
    ([], [Range(0x20, 0x40)], []),
    ([Range(0x20, 0x40)], [], [Range(0x20, 0x40)]),
    ([Range(5, 10), Range(12, 14)], [Range(0, 20)], []),
    (
        [Range(0, 30)],
        [Range(5, 10), Range(15, 20), Range(25, 30)],
        [Range(0, 5), Range(10, 15), Range(20, 25)],
    ),  # remove with gaps
    (
        [Range(0x20, 0x40), Range(0x60, 0x80)],
        [Range(0x20, 0x40), Range(0x60, 0x80)],
        [],
    ),  # remove everything, exact match
]


@pytest.mark.parametrize("test_case", REMOVE_SUBRANGES_TEST_CASES)
def test_remove_subranges(test_case):
    ranges, to_remove, expected_result = test_case
    assert remove_subranges(ranges, to_remove) == expected_result


@pytest.mark.parametrize(
    "test_range, offset, expected_range",
    [
        (Range(0, 10), 0, Range(0, 10)),
        (Range(0, 10), 5, Range(5, 15)),
    ],
)
def test_translate(test_range: Range, offset: int, expected_range: Range):
    translated_range = test_range.translate(offset)
    assert translated_range == expected_range


@pytest.mark.parametrize("test_range, offset", [(Range(5, 10), -6)])
def test_translate_value_error(test_range: Range, offset: int):
    with pytest.raises(ValueError):
        test_range.translate(offset)


def test_translate_overflow():
    """Test translate with values that would overflow."""
    r = Range(Range.MAX - 10, Range.MAX)
    with pytest.raises(OverflowError):
        r.translate(100)


def test_repr():
    test_range = Range(5, 20)
    assert test_range.__repr__() == "Range(0x5, 0x14)"


def test_hash():
    """
    Test `Range.__hash__`, which is used in lookups like below.
    """
    range_set = {Range(10, 20)}
    assert Range(10, 20) in range_set


@given(start=strategies.integers(), size=strategies.integers(min_value=0))
def test_from_size(start: int, size: int):
    range_from_size = Range.from_size(start, size)
    assert range_from_size.start == start
    assert range_from_size.length() == size


MERGE_RANGES_TEST_CASES = [
    ("no ranges", [], []),
    ("single range", [Range(0, 10)], [Range(0, 10)]),
    (
        "non-overlapping ranges",
        [Range(0, 10), Range(12, 20)],
        [Range(0, 10), Range(12, 20)],
    ),
    (
        "simple adjacent ranges",
        [Range(0, 10), Range(10, 20)],
        [
            Range(0, 20),
        ],
    ),
    (
        "simple overlapping ranges",
        [Range(0, 12), Range(10, 20)],
        [
            Range(0, 20),
        ],
    ),
    (
        "many unordered overlapping ranges",
        [Range(90, 100), Range(35, 42), Range(10, 25), Range(85, 85), Range(80, 90), Range(20, 40)],
        [Range(10, 42), Range(80, 100)],
    ),
    (
        "ranges with negative starts/ends",
        [
            Range(-100, -90),
            Range(-42, -35),
            Range(-25, -10),
            Range(-85, -85),
            Range(-90, -80),
            Range(-40, -20),
        ],
        [Range(-100, -80), Range(-42, -10)],
    ),
    (
        "many duplicated ranges",
        [Range(0, 10), Range(10, 20), Range(10, 20), Range(10, 20), Range(10, 20), Range(10, 20)],
        [
            Range(0, 20),
        ],
    ),
    (
        "very large ranges",
        [Range(0, int(Range.MAX / 2)), Range(int(Range.MAX / 2), Range.MAX)],
        [
            Range(0, Range.MAX),
        ],
    ),
]


@pytest.mark.parametrize("test_case", MERGE_RANGES_TEST_CASES, ids=lambda tc: tc[0])
def test_merge_ranges(test_case):
    _, input_ranges, expected_merged_ranges = test_case

    merged_ranges = Range.merge_ranges(input_ranges)
    assert expected_merged_ranges == merged_ranges


@dataclass
class ChunkRangesTestCase:
    ranges: List[Range]
    chunck_size: int
    expected_output: List[Range]


RANGES_CHUNK_TEST_CASES = [
    ChunkRangesTestCase([Range(0, 2)], 1, [Range(0, 1), Range(1, 2)]),
    ChunkRangesTestCase([Range(0, 4)], 2, [Range(0, 2), Range(2, 4)]),
    ChunkRangesTestCase([Range(0, 4), Range(4, 8)], 2, [Range(i, i + 2) for i in range(0, 8, 2)]),
    ChunkRangesTestCase([Range(0, 4), Range(1, 3)], 2, [Range(0, 2), Range(2, 4)]),
    ChunkRangesTestCase([Range(0, 14)], 4, [Range(0, 4), Range(4, 8), Range(8, 12), Range(12, 14)]),
]


@pytest.mark.parametrize("test_case", RANGES_CHUNK_TEST_CASES)
def test_chunk_ranges(test_case: ChunkRangesTestCase):
    ranges = chunk_ranges(test_case.ranges, test_case.chunck_size)
    assert ranges == test_case.expected_output


def test_range_iterable():
    r = Range(10, 12)
    # This should return without raising a TypeError if `r` is iterable.
    iter(r)
    # Also verify isinstance() recognizes Range as iterable
    assert isinstance(r, Iterable)


def test_range_iterable_for_loop():
    """Test that Range behaves like an iterable, notably in for loops."""
    r = Range(10, 12)
    # Test that the same object can be iterated over several times
    for iteration in range(2):
        range_iterated_items = [i for i in r]
        assert range_iterated_items == list(range(10, 12))


def test_empty_range():
    """Test behavior of empty ranges (start == end)."""
    r = Range(10, 10)
    assert r.length() == 0
    assert 10 not in r
    assert list(r) == []
    assert r.overlaps(Range(0, 20))
    assert r.within(Range(0, 20))


def test_negative_ranges():
    """Test ranges with negative values."""
    r = Range(-10, -5)
    assert r.length() == 5
    assert -7 in r
    assert -10 in r
    assert -5 not in r
    assert list(r) == [-10, -9, -8, -7, -6]


def test_very_large_ranges():
    """Test ranges near Range.MAX."""
    # Test with large values
    large_start = Range.MAX - 100
    r = Range(large_start, Range.MAX)
    assert r.length() == 100
    assert large_start in r
    assert Range.MAX - 1 in r
    assert Range.MAX not in r


def test_chunk_ranges_invalid_chunk_size():
    """Test chunk_ranges with invalid chunk sizes."""
    ranges = [Range(0, 10)]

    with pytest.raises(ValueError):
        chunk_ranges(ranges, 0)

    with pytest.raises(ValueError):
        chunk_ranges(ranges, -1)


def test_contains_constant_time():
    """Verify that __contains__ is more efficient than iteration."""
    import time

    # Create a very large range
    large_range = Range(0, 1_000_000)

    # Test __contains__ (should be constant time)
    start_time = time.perf_counter()
    for _ in range(10000):
        assert 500_000 in large_range
    contains_time = time.perf_counter() - start_time

    # Test iteration (should be much slower for middle values)
    start_time = time.perf_counter()
    for _ in range(10):  # Much fewer iterations because this is slow
        found = False
        for val in large_range:
            if val == 500_000:
                found = True
                break
        assert found
    iter_time = time.perf_counter() - start_time

    # The iteration method should be significantly slower
    # We don't assert exact ratios due to system variability
    # but __contains__ should be orders of magnitude faster
    assert contains_time < iter_time
